package com.nisum.inventory.service;

import com.mongodb.MongoClientSettings;
import com.mongodb.client.MongoClient;
import com.mongodb.client.MongoClients;
import com.mongodb.client.MongoCollection;
import com.mongodb.client.MongoDatabase;
import com.mongodb.client.result.InsertOneResult;
import com.nisum.inventory.collection.ProcessedInventory;
import com.nisum.inventory.dto.*;
import com.nisum.inventory.repository.InventoryMapper;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.spark.sql.*;
import org.apache.spark.sql.execution.streaming.sources.ForeachWriterTable;
import org.apache.spark.sql.streaming.OutputMode;
import org.apache.spark.sql.streaming.StreamingQuery;
import org.apache.spark.sql.streaming.StreamingQueryException;
import org.apache.spark.sql.streaming.Trigger;
import org.apache.spark.sql.types.*;
import org.bson.codecs.configuration.CodecRegistries;
import org.bson.codecs.configuration.CodecRegistry;
import org.bson.codecs.pojo.PojoCodecProvider;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.io.buffer.DataBufferUtils;
import org.springframework.data.mongodb.core.ReactiveMongoOperations;
import org.springframework.http.codec.multipart.FilePart;
import org.springframework.stereotype.Service;
import reactor.core.publisher.Flux;

import java.io.File;
import java.io.IOException;
import java.io.Serializable;
import java.nio.charset.StandardCharsets;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.concurrent.TimeoutException;

@Slf4j
@Service
public class InventoryService implements Serializable {


    @Autowired
    private SparkSession sparkSession;

    @Autowired
    private StructType schema;


    public Flux<KafkaPublisherResponse> saveAndProcess(Flux<FilePart> filePartFlux) {

        return filePartFlux.flatMap(filePart ->
                filePart.content().map(dataBuffer -> {
                    byte[] bytes = new byte[dataBuffer.capacity()];
                    dataBuffer.read(bytes);
                    DataBufferUtils.release(dataBuffer);

                    return FileUploadAttributes.builder()
                            .fileContent(new String(bytes, StandardCharsets.UTF_8))
                            .fileName(filePart.filename())
                            .build();

                })
                        .map(content ->  saveAsFile(content))
                        .map(fileName -> readAndPublishDataToKafka(fileName))
                        .map(kafkaPublisherResponse -> kafkaPublisherResponse));
    }

    private String saveAsFile(FileUploadAttributes fileUploadAttributes) {

        try {
            FileUtils.deleteDirectory(new File("src/main/resources/data/CheckPointLocation"));
        } catch (IOException e) {
            log.error("Error while deleting checkpoint location dir {}", e.getMessage());
        }
        String batchId= new SimpleDateFormat("HHmmssSSS").format(new Date());
        String dir="src/main/resources/data/"+batchId;
        String fileName = dir+ File.separator + fileUploadAttributes.getFileName();
        try {
            FileUtils.forceMkdir(new File(dir));
            FileUtils.writeStringToFile(new File(fileName)
                    , fileUploadAttributes.getFileContent(), StandardCharsets.UTF_8, true);

            return batchId;
        } catch (IOException e) {
            if (e instanceof IOException)
                log.error("Error while creating CSV... {}", e.getMessage());
            return StringUtils.EMPTY;
        }

    }

    private KafkaPublisherResponse readAndPublishDataToKafka(String batchId) {

        if (StringUtils.isNotEmpty(batchId)) {
            String topicName = "test-inventory2";
            try {
                populateInventoryDataUsingDFStruct(batchId, topicName);
                return KafkaPublisherResponse.builder()
                        .topicName(topicName)
                        .batchId(batchId)
                        .status("File data pushed to Kafka topic.").build();
            } catch (TimeoutException | StreamingQueryException e) {
                if (e instanceof TimeoutException)
                    log.error("TimeoutException... {}", e.getMessage());
                if (e instanceof StreamingQueryException)
                    log.error("StreamingQueryException  {}", e.getMessage());
                return KafkaPublisherResponse.builder().topicName(topicName).status(e.getMessage()).build();
            }
        }else
        {
            return KafkaPublisherResponse.builder().build();
        }
    }



    public void populateInventoryDataUsingDFStruct(String batchId,String topicName) throws TimeoutException, StreamingQueryException {


        StructType structType = Encoders.bean(Root.class).schema();

        Dataset<Row> df=sparkSession.readStream()
                .format("json")
                .option("multiline",true)
                .schema(structType)
                .load("src/main/resources/data/"+batchId);

        //df.printSchema();



        Dataset<Row> rowDataset=df
                .withColumn("uniqueId", functions.lit(UUID.randomUUID().toString()))
                .withColumn("batchId",functions.lit(batchId));

        rowDataset.selectExpr("CAST(batchId AS STRING) AS key", "to_json(struct(*)) AS value")
                .writeStream()
                .format("kafka")
                .trigger(Trigger.Once())
                .option("kafka.bootstrap.servers", "HYD-LAP-00484.corp.nisum.com:9092")
                .option("topic", topicName)
                .option("checkpointLocation", "src/main/resources/data/CheckPointLocation")
                .start()
                .awaitTermination(5000);


    }


    public Flux<Root> getProcessedInventory(String batchId) throws TimeoutException, StreamingQueryException {


        Dataset<Row> df = sparkSession
                .readStream()
                .format("kafka")
                .option("kafka.bootstrap.servers", "HYD-LAP-00484.corp.nisum.com:9092")
                .option("subscribe", "test-inventory2")
                .option("startingOffsets", "earliest")
                .option("startingOffsets", "earliest")
                .option("failOnDataLoss", "false")
                .option("multiline",true)
                .option("kafka.con.group.id", "localStream")
                .load();

        StructType schema = Encoders.bean(Root.class).schema();

        df.printSchema();

        Dataset<RawInventory> inventoryList=df
                .filter(functions.col("key").equalTo(batchId))
                .selectExpr("CAST(value as String) as message")
                .select(functions.from_json(functions.col("message"),schema ).as("data"))
                .select(functions.explode(functions.col("data.inventoryList")).as("list"))
                .filter(
                            functions.col("list.price").leq(15)
                            .and(functions.col("list.itemType").contains("mug"))
                            .and(functions.col("list.name").contains("Coffee"))
                )
                .select("list.*")
                .as(Encoders.bean(RawInventory.class));


        StreamingQuery streamingQuery=inventoryList
                .writeStream()
                .trigger(Trigger.Once())
                .foreach(new ForeachWriter<RawInventory>() {

                    private MongoCollection<ProcessedInventory> processedInventoryMongoCollection;
                    private MongoClient client;

                    @Override
                    public boolean open(long partitionId, long epochId) {
                        String uri = "mongodb://127.0.0.1:27017";
                        client = MongoClients.create(uri);

                        CodecRegistry pojoCodecRegistry = CodecRegistries.fromRegistries(
                                MongoClientSettings.getDefaultCodecRegistry(),
                                CodecRegistries.fromProviders(PojoCodecProvider.builder().automatic(true).build()));

                        // Accessing the database
                        MongoDatabase mongoDatabase = client.getDatabase("inventorydb").withCodecRegistry(pojoCodecRegistry);
                        processedInventoryMongoCollection = mongoDatabase.getCollection("inventory",ProcessedInventory.class);
                        return true;
                    }

                    @Override
                    public void process(RawInventory value) {

                        InsertOneResult insertOneResult=processedInventoryMongoCollection.insertOne(InventoryMapper.INSTANCE.dtoToEntity(value,UUID.randomUUID().toString()));
                        log.info("Inserted Result : {} ",insertOneResult);
                    }

                    @Override
                    public void close(Throwable errorOrNull) {
                        client.close();
                    }
                }).start();

       /*StreamingQuery streamingQuery=inventoryList
                .writeStream()
                .trigger(Trigger.Once())
                .format("json")
                .option("multiline",true)
                .option("checkpointLocation", "src/main/resources/data/checkpointLocation/StreamingJob")
                .option("path","src/main/resources/data/"+batchId)
                .outputMode(OutputMode.Append())
                .start();*/

        streamingQuery.awaitTermination();

        return Flux.empty();

    }



}
